/*
  ochimono_AQM1248A.ino Ver.1.00
  
  ochimono is a demonstrational sketch for AQM1248A LCD module
  It shows how to make a game sketch using MGLCD library.
  
  ochimono was originally developed for 122X32 MGLCD shield by Synapse
(Hiroshi Tanigawa) in 2013.  And it was ported to AQM1248A by himself in
2016.
  This sketch is originally distributed at Team Schaft's Homepage.
  <http://www3.big.or.jp/~schaft/hardware/>

  ochimono is free software: you can redistribute it and/or modify
  it under the terms of the GNU Lesser General Public License as published by
  the Free Software Foundation, either version 2.1 of the License, or
  (at your option) any later version.

  ochimono is distributed in the hope that it will be useful,
  but WITHOUT ANY WARRANTY; without even the implied warranty of
  MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
  GNU Lesser General Public License for more details.

  You should have received a copy of the GNU Lesser General Public License
  along with MGLCD.  If not, see <http://www.gnu.org/licenses/>.
  
*/

#include <MGLCD.h>

// ハードウェアSPIの場合は以下の2行を有効にする(Arduino IDE 1.6.Xまたは1.7.Xが必要。他にも変更点があるので注意)
// #include <SPI.h>
// #include <MGLCD_SPI.h>

#include "num0.h"
#include "num1.h"
#include "num2.h"
#include "num3.h"
#include "num4.h"
#include "num5.h"
#include "num6.h"
#include "num7.h"
#include "num8.h"
#include "num9.h"
#include "title.h"
#include "start.h"
#include "lines.h"
#include "kyoto.h"
#include "message.h"
#include "gameover.h"

const int HorizBlockNum   =   11; // 水平方向のブロック数
const int VertBlockNum    =   21; // 垂直方向のブロック数
const int HorizOffset     =   45; // 水平方向のブロック表示位置オフセット
const int VertOffset      =  125; // 垂直方向のブロック表示位置オフセット
const int PieceTypeNum    =    7; // ピースの種類の数
const int BlockPitch      =    4; // ブロックが並ぶピッチ
const int KeyMillis       =   30; // キーを読む間隔(ms)
const int InitialFallTime =   30; // 最初の自動落下時間(* KeyMills ms)
const int RepeatTime      =   10; // キーリピートの時間(* KeyMills ms)
const int FallRepeatTime  =    2; // 落下ボタンを押した場合のキーリピート時間(* KeyMills ms)
const int InvalidKey      =   -1; // キーを押していなかった

const int LeftKey         =    0; // 左ボタン  (SW1)
const int RightKey        =    1; // 右ボタン  (SW2)
const int RotateKey       =    2; // 回転ボタン(SW3)
const int FallKey         =    3; // 落下ボタン(SW4)

const int SpeedUpInterval = 5000; // 自動落下時間を短くする間隔(* KeyMills ms)

#define CS_PIN   10
#define DI_PIN    9

// ハードウェアSPIの場合は以下の3行を無効にする
#define SCK_PIN  13
#define MOSI_PIN 11
static MGLCD_AQM1248A_SoftwareSPI MGLCD(MGLCD_SpiPin4(SCK_PIN, MOSI_PIN, CS_PIN, DI_PIN));

// ハードウェアSPIの場合は以下の以下の2行を有効にする
// #define MAX_FREQ (1000*1000UL)
// static MGLCD_AQM1248A_SPI MGLCD(MGLCD_SpiPin2(CS_PIN, DI_PIN), MAX_FREQ);

struct TPosition { // 2次元の位置・ベクトル
  int x; // 水平座標
  int y; // 垂直座標
  
  TPosition Rotate(byte n=1); // 原点を中心にn回右回りに90°回転した時の座標
  boolean BlockExists(); // その位置にブロックがあるかどうか
  void Draw(byte color=1); // その位置にブロックを描く
  TPosition operator+(const TPosition &pos); // 2次元ベクトルの足し算
  boolean operator==(const TPosition &pos); // 位置が一致するかどうか
  boolean operator!=(const TPosition &pos); // 位置が一致しないかどうか
}; // TPosition

TPosition TPosition::Rotate(byte n)
{
  TPosition result;
  
  result.x=x;
  result.y=y;
  
  for(int i=0; i<n; i++) {
    int xx=result.x;
    int yy=result.y;
    
    result.x=-yy;
    result.y=xx;
  } // for i
  
  return result;
} // TPosition::Rotate

boolean TPosition::BlockExists()
{
  return MGLCD.GetPixel(VertOffset-y*BlockPitch,HorizOffset-x*BlockPitch)!=0;
} // TPosition::BlockExists

TPosition TPosition::operator+(const TPosition &pos)
{
  TPosition result;
  
  result.x=x+pos.x;
  result.y=y+pos.y;
  
  return result;
} // TPosition::operator+

void TPosition::Draw(byte color)
{
  MGLCD.FillRect(VertOffset -(y  )*BlockPitch  ,
                 HorizOffset-(x  )*BlockPitch  ,
                 VertOffset -(y+1)*BlockPitch+2,
                 HorizOffset-(x+1)*BlockPitch+2,
                 color);
} // TPosition::Draw

boolean TPosition::operator==(const TPosition &pos)
{
  return x==pos.x && y==pos.y;
} // TPosition::operator==

boolean TPosition::operator!=(const TPosition &pos)
{
  return x!=pos.x || y!=pos.y;
} // TPosition::operator!=

struct TPieceDef{ // 1種類のピースの定義
  TPosition pos[4]; // 4つのブロックの相対位置
  int RotationNum; // 回転できる回数
}; // TPieceDef

const TPieceDef PieceDef[PieceTypeNum]= { // 7種類のピースの定義
  {{{-1, 0},{ 0, 0},{ 1, 0},{ 2, 0}},2}, // 0 直線
  {{{-1, 1},{ 0, 1},{-1, 0},{ 0, 0}},1}, // 1 田
  {{{-1, 0},{ 0, 0},{ 0, 1},{ 1, 1}},2}, // 2 鍵
  {{{-1, 1},{ 0, 1},{ 0, 0},{ 1, 0}},2}, // 3 逆鍵
  {{{ 0, 1},{ 0, 0},{ 1, 0},{ 2, 0}},4}, // 4 逆L
  {{{-2, 0},{-1, 0},{ 0, 0},{ 0, 1}},4}, // 5 L
  {{{-1, 0},{ 0, 1},{ 1, 0},{ 0, 0}},4}  // 6 T
}; // PieceDef 

struct TPiece { // ピースの種類、位置、回転方向の情報
  int type; // 種類
  TPosition pos; // 位置
  int rotation; // 回転情報
  
  boolean CanPlace(TPiece &CurrentPiece); // ピースを描画できるか調べる
  void Draw(byte color=1); // ピースを描画する
  boolean Fall(); // ピースを1段落とす
  boolean GetNew(); // 新規の落下ピースを取得する
}; // TPiece

boolean TPiece::CanPlace(TPiece &CurrentPiece)
{
  TPosition CurrentPos[4];
  
  if((&CurrentPiece)!=NULL) {
    for(int i=0; i<4; i++) {
      TPosition p=PieceDef[CurrentPiece.type].pos[i];
      CurrentPos[i]=CurrentPiece.pos+p.Rotate(CurrentPiece.rotation);
    } // int i
  } // if
  
  for(int i=0; i<4; i++) {
    TPosition p=PieceDef[type].pos[i];
    p=pos+p.Rotate(rotation);
    if (p.x<0 || p.x>=HorizBlockNum || p.y<0 || p.y>=VertBlockNum) return false; // ブロックが画面外にある
    
    boolean found = false;
    if((&CurrentPiece)!=NULL) {
      for(int j=0; j<4; j++) {
        if(p==CurrentPos[j]) {
          found=true;
          break;
        } // if
      } // for j
    } // if
    
    if(!found && p.BlockExists()) return false; // ブロックが既存のブロックと重なっている
  } // int i
  return true;
} // TPiece::CanPlace

void TPiece::Draw(byte color)
{
  for(int i=0; i<4; i++) {
    TPosition p=PieceDef[type].pos[i];
    (pos+p.Rotate(rotation)).Draw(color);    
  } // for i
} // TPiece::Draw

boolean TPiece::Fall()
{
  TPiece NextPiece=*this;
  NextPiece.pos.y--;
  if(!NextPiece.CanPlace(*this)) return false;
  this->Draw(0);
  *this=NextPiece;
  this->Draw(1);
  return true;
} // TPiece::Fall

boolean TPiece::GetNew()
{
  type=random(PieceTypeNum);
  pos.x=random(HorizBlockNum-4)+2;
  pos.y=VertBlockNum-3;
  rotation=random(PieceDef[type].RotationNum);
  
  return CanPlace(*(TPiece *)NULL);
} // TPiece::GetNew

struct TPosition Position(int x, int y)
{
  TPosition result;
  result.x=x;
  result.y=y;
  return result;
} // Position

// 枠を描く
void DrawFrame()
{
  // タイトルの描画
  MGLCD.DrawBitmap(BMP_title,0,0,MGLCD_ROM);
  
  // "LINES"の描画
  MGLCD.DrawBitmap(BMP_lines,18,10,MGLCD_ROM);

  // 枠の描画
  MGLCD.Rect(VertOffset+2,HorizOffset+2,VertOffset-VertBlockNum*BlockPitch,HorizOffset-HorizBlockNum*BlockPitch);
} // DrawFrame

// スコアを画面に表示する
void WriteScore(unsigned score)
{
  const int HorizOffset= 5;
  const int VertOffset = 30;
  PROGMEM const byte *p;
  int x=HorizOffset;
  
  do {
    switch(score%10) {
      case 0:
        p=BMP_num0;
        break;
      case 1:
        p=BMP_num1;
        break;
      case 2:
        p=BMP_num2;
        break;
      case 3:
        p=BMP_num3;
        break;
      case 4:
        p=BMP_num4;
        break;
      case 5:
        p=BMP_num5;
        break;
      case 6:
        p=BMP_num6;
        break;
      case 7:
        p=BMP_num7;
        break;
      case 8:
        p=BMP_num8;
        break;
      case 9:
        p=BMP_num9;
        break;
    } // switch
    MGLCD.DrawBitmap(p,VertOffset,x,MGLCD_ROM);
    x+=6;
    score/=10;
  } while(score!=0);
} // WriteScore

static int8_t ReadButton(uint8_t n)
{
  const uint8_t pin[4]={8,7,4,2};
  
  if(digitalRead(pin[n])==HIGH) {
    return MGLCD_BUTTON_OFF;
  } else {
    return MGLCD_BUTTON_ON;
  } // if
} // ReadButton

// 横一列に揃った行をつぶす。返り値は横一列に揃った行の数。
int CrushLines()
{
  int result=0;
  
  for(int y=VertBlockNum-1; y>=0; y--) {
    boolean flag=true;
    while(flag) {
      // 1行揃っているか調べる
      for(int x=0; x<HorizBlockNum; x++) {
        if(!Position(x,y).BlockExists()) {
          flag=false;
          break;
        } // if
      } // for x
      
      if(flag) { // 1行揃っていた
        for(int yy=y; yy<VertBlockNum-1; yy++) {
          for(int x=0; x<HorizBlockNum; x++) {
            Position(x,yy).Draw(Position(x,yy+1).BlockExists() ? 1 : 0);
          } // for x
        } // for yy
        result++;
      } // if
    } // while
  } // for
   
  return result; 
} // CrushLines

// オープニング時の画面表示
void opening()
{
  // タイトルの表示
  MGLCD.DrawBitmap(BMP_title,0,0,MGLCD_ROM);
  
  delay(500);
  
  // 京都のロゴを表示
  for(int i=18; i<=99; i++) {
    delay(20);
    MGLCD.Line(i-1,0,i-1,31,0);
    MGLCD.DrawBitmap(BMP_kyoto,i,8,MGLCD_ROM);
  } // for i
  
  // "This software was..."を表示
  for(int i=0; i<=10; i++) {
    for(int j=0; j<3; j++) {
      MGLCD.FillRect(43,0,93,47,0);
      delay((10-i)*7);
      MGLCD.DrawBitmap(BMP_message,43,0,MGLCD_ROM);
      delay(i*7);
    } // for j
  } // for i
  
  // タイトルの表示
  for(int i=1; i<=20; i++) {
    delay(60);
    MGLCD.Line(i-1,0,i-1,47,0);
    MGLCD.DrawBitmap(BMP_title,i,0,MGLCD_ROM);
  }

  delay(5000);
} // opening

// ゲームオーバー時の画面表示
void GameOver()
{
  // 盤面の反転
  for(int y=VertOffset+1; y>VertOffset-BlockPitch*VertBlockNum; y--) {
    for(int x=HorizOffset+1; x>HorizOffset-BlockPitch*HorizBlockNum; x--) {
      MGLCD.SetPixel(y,x,!MGLCD.GetPixel(y,x));
    } // for x
    delay(30);
  } // for y
  
  // "GAME OVER"の表示
  MGLCD.FillRect(45,8,84,39,0);
  MGLCD.DrawBitmap(BMP_gameover,53,8,MGLCD_ROM);
} // GameOver

// キー入力を待つ
void WaitForKey()
{
  int LastKey;
  int key=InvalidKey;
  
  // キーの初期状態を取得
  for(int i=0; i<4; i++) {
    if(ReadButton(i)==MGLCD_BUTTON_ON) {
      key=i;
      break;
    } // if
  } // for i
  
  // キーの状態が変わるのを待つ
  do {
    LastKey=key;
    key=InvalidKey;
    for(int i=0; i<4; i++) {
      if(ReadButton(i)==MGLCD_BUTTON_ON) {
        key=i;
        break;
      } // if
    } // for i  
  } while(key!=InvalidKey|| key==LastKey);
} // WaitForKey

void setup()
{
  pinMode( 0,INPUT_PULLUP); // not used
  pinMode( 1,INPUT_PULLUP); // not used
  pinMode( 2,INPUT_PULLUP); // SW1
  pinMode( 3,INPUT_PULLUP); // not used
  pinMode( 4,INPUT_PULLUP); // SW2
  pinMode( 5,INPUT_PULLUP); // not used
  pinMode( 6,INPUT_PULLUP); // not used
  pinMode( 7,INPUT_PULLUP); // SW3
  pinMode( 8,INPUT_PULLUP); // SW4
  pinMode( 9,OUTPUT      ); // DI
  pinMode(10,OUTPUT      ); // CS
  pinMode(11,OUTPUT      ); // MOSI
  pinMode(12,INPUT       ); // MISO
  pinMode(13,OUTPUT      ); // SCK
  pinMode(A0,INPUT_PULLUP); // not used
  pinMode(A1,INPUT_PULLUP); // not used
  pinMode(A2,INPUT_PULLUP); // not used
  pinMode(A3,INPUT_PULLUP); // not used
  pinMode(A4,INPUT_PULLUP); // not used
  pinMode(A5,INPUT_PULLUP); // not used
  
  while(MGLCD.Reset()); // LCD初期化
  opening();
} // setup

void loop()
{
  TPiece CurrentPiece; // 現在落下中のピース
  int FallTime=InitialFallTime; // 自動落下時間(* KeyMills ms)
  unsigned score=0; // スコア
    
  MGLCD.ClearScreen(); // 画面を消去  
  MGLCD.DrawBitmap(BMP_start,22,2,MGLCD_ROM); // "PRESS ANY KEY TO START"を表示

  WaitForKey(); // キー入力を待つ
  MGLCD.ClearScreen(); // 画面消去
  DrawFrame(); // フレームを書く
  WriteScore(score); // スコア表示
  CurrentPiece.GetNew(); // 新規のピースを取得する
  CurrentPiece.Draw(); // 新規のピースを描く
  
  unsigned long LastMillis=millis(); // 経過時間計測用
  int LastKey=InvalidKey; // 1回前に押していたキー
  int RepeatCnt=0; // キーリピート用カウンタ(落下以外)
  int FallRepeatCnt=0; // キーリピート用カウンタ(落下)
  int FallCnt=0; // 自動落下用カウンタ
  int SpeedUpCnt=0; // 自動落下高速化用カウンタ
  while(true) {  
    if(millis()-LastMillis>KeyMillis) { // キーを読むタイミングが来た
      // キーを読む
      int key=InvalidKey;
      for(int i=0; i<4; i++) {
        if(ReadButton(i)==MGLCD_BUTTON_ON) {
          key=i;
          break;
        } // if
      } // for i;
      
      RepeatCnt++;
      FallRepeatCnt++;
      if(key!=InvalidKey && (key!=LastKey || RepeatCnt>=RepeatTime || (key==FallKey && FallRepeatCnt>=FallRepeatTime))) { // キーが押された、またはキーリピートした
        TPiece NextPiece=CurrentPiece;
        switch(key) {
          case LeftKey: // 左
            NextPiece.pos.x--;
            break;
          case RightKey: // 右
            NextPiece.pos.x++;
            break;
          case RotateKey: // 回転
            NextPiece.rotation=(NextPiece.rotation+1)%PieceDef[NextPiece.type].RotationNum;
            break;
          case FallKey: // 落下
            NextPiece.pos.y--;
            break;
          default:
            break;
        } // switch

        if((NextPiece.pos!=CurrentPiece.pos || NextPiece.rotation!=CurrentPiece.rotation) && NextPiece.CanPlace(CurrentPiece)) { // ピースが動かせた      
          CurrentPiece.Draw(0); // 古いピースを消す
          NextPiece.Draw(1); // ピースを更新する
          CurrentPiece=NextPiece; // 新しいピースを書く
          if(key==FallKey) FallCnt=0; // 自分で落下した場合は、自動落下を遅らせる
        } // if
        
        RepeatCnt=FallRepeatCnt=0; // キーリピート用カウンタのリセット
      } // if
      
      if(++FallCnt>=FallTime) { // 自動落下する時間がやってきた
        if(!CurrentPiece.Fall()) { // 落下させられなかった    
          score+=CrushLines(); // 揃った行を消す 
          WriteScore(score); // スコア表示  
          if(!CurrentPiece.GetNew()) break; // 新規のピースを取得できなかったら、ゲームオーバー
        } // if
        
        FallCnt=0; // 自動落下用カウンタのリセット
      } // if
   
      if(++SpeedUpCnt>=SpeedUpInterval) { // 落下速度を早くする
        if(--FallTime<1) FallTime=1;
        SpeedUpCnt=0;
      } // if        
   
      LastKey=key;
      LastMillis+=KeyMillis;
    } // if
  } // while
  
  GameOver();
  delay(5000);
} // loop


